
//****************************************************************************
//  ws2812 - Serial driver for ws2812 RGB leds using fastpwm
//
//  12.12.2012 - M. Marquardt (adrock0905@alice.de)
//

//----- Includes

#include <avr/io.h>
#include "asmdefs.h"


//----- Constants

//#define    F_CPU   16000000           // should already be defined somewhere
#define     F_OUT       800000                  // Output frequency
#define     CNT_TOTAL   ((F_CPU/F_OUT)-1)       // # cycles for one bit
#define     CNT_LOW     ((F_CPU/F_OUT)/3-1)     // # cycles for T0H
#define     CNT_HIGH    (((F_CPU/F_OUT)*2)/3-1) // # cycles for T1H


//----- Functions
//
//      GCC style parameter register usage:
//      r0: scratch
//      r1: always zero
//      r2-r17, r28-r29: leave unchanged/saved
//      r18-r27, r30-r31: scratch
//      Function arguments - allocated left to right, r25 to r8.
//      All arguments are aligned to start in even-numbered registers
//      (odd-sized arguments, including char, have one free register above
//      them).
//      Return values: 8-bit in r24 (not r25!), 16-bit in r25:r24


//----- Some macros used later

//  BITOUT - check for bitnum set in reg and set OCR0B val accordingly

.macro  BITOUT reg, bitnum
    STORE   TIFR, r22                   // clear OCF0A flag

    STORE   OCR0B, r18                  // set OCR0B to low val
    sbrc    \reg, \bitnum               // skip if bit is cleared
    STORE   OCR0B, r19                  // set OCR0B to high val
.endm

//  BITWAIT - wait for actual pwm cycle to finish

.macro  BITWAIT
1:
    LOAD    r20, TIFR
    sbrs    r20, OCF0A
    rjmp    1b
.endm


//****************************************************************************
//  void WS_init(void)
//
//      Initialize hardware
//
.global WS_init

WS_init:
    cbi     _SFR_IO_ADDR(PORTB), PB1
    sbi     _SFR_IO_ADDR(DDRB), PB1
    ldi     r24, (1<<COM0B1)|(1<<WGM01)|(1<<WGM00)
    STORE   TCCR0A, r24
    ldi     r24, CNT_TOTAL
    STORE   OCR0A, r24
   
    ret


//****************************************************************************
//  void WS_put(uint8_t *data, uint16_t count)
//
//      Output bytes to ws2812 leds using fastpwm.d.
//
//  IN:
//      r24: uint8_t *data      Ptr. to bytes to output
//      r22: uint16_t count     Number of bytes
//
.global WS_put

WS_put:
    movw    r26, r24                    // X = Ptr to output data
    movw    r24, r22                    // r24/r25 = Output counter

    ldi     r22, (1<<OCF0A)             // r22 = value to reset OCF0A
    ldi     r23, (1<<CS00)|(1<<WGM02)   // r23 = value to start counter
    ldi     r18, CNT_LOW                // r18 = low OCR0B value
    ldi     r19, CNT_HIGH               // r19 = high OCR0B value

    STORE   OCR0B, r1                   // OCR0B = 0 -> zero cycle at beginning
    STORE   TCNT0, r1                   // Clear timer
    STORE   TCCR0B, r23                 // start timer

    // X: ptr to next byte, r24/r25: Output byte counter

WS_put1:
    sbiw    r24, 0x01                   // dec counter
    brcs    WS_put_end

    ld      r0, X+                      // Get next byte

    BITOUT  r0, 7
    BITWAIT
    BITOUT  r0, 6
    BITWAIT
    BITOUT  r0, 5
    BITWAIT
    BITOUT  r0, 4
    BITWAIT
    BITOUT  r0, 3
    BITWAIT
    BITOUT  r0, 2
    BITWAIT
    BITOUT  r0, 1
    BITWAIT
    BITOUT  r0, 0
    BITWAIT

    rjmp    WS_put1

WS_put_end:
    STORE   TIFR, r22                   // clear OCF0A flag
    STORE   OCR0B, r1                   // last OCR0B = 0 

WS_put_waitfinal:
    LOAD    r20, TIFR
    sbrs    r20, OCF0A
    rjmp    WS_put_waitfinal

    STORE   TCCR0B, r1                  // stop timer
    ret


//****************************************************************************
//  void WS_putG(uint8_t *data, uint16_t count, uint8_t *gamma)
//
//      Output bytes to ws2812 leds using fastpwm with gamma correction using
//      the lookup table specified.
//      Due to implementation constraints, there is ONE MORE byte read at end
//      of data, but not output.
//
//  IN:
//      r24: uint8_t *data      Ptr. to bytes to output
//      r22: uint16_t count     Number of bytes
//      r20: uint8_t *gamma     Ptr. to GammaTable in FLASH
//
.global WS_putG

WS_putG:
    push    r28
    push    r29
    movw    r28, r20                    // Y = Ptr. to gamme table

    movw    r26, r24                    // X = Ptr to output data
    movw    r24, r22                    // r24/r25 = Output counter

    ldi     r22, (1<<OCF0A)             // r22 = value to reset OCF0A
    ldi     r23, (1<<CS00)|(1<<WGM02)   // r23 = value to start counter
    ldi     r18, CNT_LOW                // r18 = low OCR0B value
    ldi     r19, CNT_HIGH               // r19 = high OCR0B value

    STORE   OCR0B, r1                   // OCR0B = 0 -> zero cycle at beginning
    STORE   TCNT0, r1                   // Clear timer
    STORE   TCCR0B, r23                 // start timer

    // X: ptr to next byte, r24/r25: Output byte counter

    ld      r0, X+                      // Get next byte
    movw    r30, r28                    // gamma table into Z
    add     r30, r0                     // add byte as offset
    adc     r31, r1
    lpm     r0, Z                       // lookup byte -> r0

WS_putG1:
    sbiw    r24, 0x01                   // dec counter
    brcs    WS_putG_end

    // Some instructions for gamma table lookup for next byte are spread
    // between bit outputs for timing reasons

    BITOUT  r0, 7
    BITWAIT
    BITOUT  r0, 6
    BITWAIT
    BITOUT  r0, 5
    BITWAIT
    BITOUT  r0, 4
    BITWAIT
    BITOUT  r0, 3
    ld      r21, X+                     // Get next byte
    BITWAIT
    BITOUT  r0, 2
    movw    r30, r28                    // gamma table into Z
    BITWAIT
    BITOUT  r0, 1
    add     r30, r21                    // add byte as offset
    adc     r31, r1
    BITWAIT
    BITOUT  r0, 0
    lpm     r0, Z                       // lookup byte -> r0
    BITWAIT

    rjmp    WS_putG1

WS_putG_end:
    STORE   TIFR, r22                   // clear OCF0A flag
    STORE   OCR0B, r1                   // last OCR0B = 0 

WS_putG_waitfinal:
    LOAD    r20, TIFR
    sbrs    r20, OCF0A
    rjmp    WS_putG_waitfinal

    STORE   TCCR0B, r1                  // stop timer

    pop     r29
    pop     r28
    ret